#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module is specific to quantfit application.
It takes in all input files and processes them to the
correct format which is unbreakable inside the Quant package.
"""


from builtins import str, object
import json
import os
import ast
import re
import warnings as wrngs

from datum import algorithms as dat_alg
from datum import helpers as dat_hlp
from olmonk import ConfigDataValidator as CDV
import normalisation as nl
import numpy as np
import pandas as pd
from quant.table_processing import TableProcessing
from sklearn.metrics import silhouette_score
from sklearn.cluster import KMeans

from datum import constants
from algowrap import err_constants as err_cnsts
from datum import metadata_auto_generation as meta_gen
from datum.file_validation_checks import validate_via_olmonk


# Can we make this Exception specific?
def df_csv_validator(check_dict, path):
    """
    Calls the OlMonk CDV validator to validate the given dataframe and
    returns the logs from CDV

    Parameters
    ----------
    dict: Dict
        Dict to validate

    Returns
    -------
    df: Pandas Dataframe
        Validated df
    logs: Dict
        Dict containing the CDV logs
    """
    try:
        mq_dict = dict(check_dict)
        mq_dict['file_path'] = path
        cdv = CDV(mq_dict)
        cdv.validate()
        return cdv.dv.corrected_df, cdv.dv.logs
    except Exception as e:
        logs = {"errors": [str(e)], "warnings": {"action": [],
                                                    "message": []
                                                    }
                }
        return pd.DataFrame(), logs


def read_raw_file(path):
    """
    This function read raw file and return  df in format required for LCMS
    or LCMS/MS

    :param path(string): path of file
    :return: Dataframe having raw data and logs in the form of dictionary
    """
    try:
        file_type = dat_alg.get_file_type(path)
        if file_type == constants.MS:
            raw_file, logs = dat_alg.convert_maven_to_required_df(path,
                                                                  constants.QUANT_LCMS)
        elif file_type == constants.ISOCORRECT:
            raw_file = dat_alg.convert_polly_to_required_df(path,
                                                        constants.ISOCORRECT_QUANT_LCMS)
            logs = {constants.ERRORS: [], constants.WARNINGS: {constants.MESSAGE: [], constants.ACTION: []}}
        else:
            raw_file, logs = dat_alg.convert_maven_to_required_df(path,
                                                           constants.QUANT_MSMS)
    except:
        raw_file, logs = validate_via_olmonk(constants.DICT_READ_MULTIQUANT, path)

    return raw_file, logs


class ProcessMqDf(object):
    """
    Process the dataframe generated from multiquant files and metadata files
    to get biosample part and the standard part of the dataframe

    Attributes:
        raw_mq: dataframe of merged raw multiquant files
        metadata_std: metadata dataframe for sample and conc mapping
        metadata_mq: metadata dataframe for fragments and unlabeled fragment
                    mapping
    """

    def __init__(self, raw_mq_df, metadata_std_sample_df, metadata_mq_df):
        """
        merge mq dataframe with metadatas

        Args:
            raw_mq_df: dataframe of merged raw multiquant files
            metadata_std_sample_df: metadata dataframe for sample and conc
                                    mapping
            metadata_mq_df: metadata dataframe for fragments and unlabeled
                            fragment mapping
        """
        self.raw_mq = raw_mq_df
        self.metadata_std = metadata_std_sample_df
        self.metadata_mq = metadata_mq_df
        self._merge_mq_metadata()

    def get_standard_biosample_data(self):
        """
        Divide raw multiquant dataframe into two dataframes, one containing
        standard samples, other containing experimental samples (biosamples)

        Returns:
            standard_df: dataframe containing standard samples
            biosamples_df: dataframe containing biosamples
        """
        standard_df, biosamples_df = self._split_std_bio()
        standard_df = self._merge_std_df_metadata(standard_df)
        columns = [constants.MQ_INTENSITY, constants.MQ_CONCENTRATION]

        # set presicion
        biosamples_df[columns] = biosamples_df[columns].applymap(
            lambda x: np.round(x, decimals=constants.PRESICION))

        return standard_df, biosamples_df


    def check_std_samples_in_rawfile(self):
        """
        Check for samples in standard metadata file for each unlabeled fragment, if missing \
        then add those samples in raw file with Area value 0.
        """
        # Should use .get()
        std_samples_list = self.metadata_std.get(constants.PARENT) + "+" + self.metadata_std.get(constants.MQ_SAMPLE_NAME)\
            + "+" + self.metadata_std.get(constants.MQ_COHORT_NAME)
        raw_file_list = self.raw_mq.get(constants.PARENT) + "+" + self.raw_mq.get(constants.MQ_SAMPLE_NAME)\
            + "+" + self.raw_mq.get(constants.MQ_COHORT_NAME)
        missing_samples = set(std_samples_list) - set(raw_file_list)

        missing_data_list = []
        msgs_list = []

        for i in list(missing_samples):
            metabolite, sample, cohort = i.split("+")
            missing_data_list.append({constants.PARENT: metabolite,
                                      constants.MQ_SAMPLE_NAME: sample,
                                      constants.MQ_INTENSITY: 0,
                                      constants.MQ_COHORT_NAME: cohort,
                                      constants.MQ_FRAGMENT: metabolite})
        for a in missing_data_list:
            self.raw_mq = self.raw_mq.append(a, ignore_index=True)

        if len(missing_data_list):
            error_msg = "Unlabelled fragment(s) {} ".format(', '.join(str(data[constants.PARENT])for data in missing_data_list))\
                + "in respective sample(s) {} ".format(', '.join(str(data[constants.MQ_SAMPLE_NAME])for data in missing_data_list))\
                 + "in your standard sample metadata are missing in the raw intensity file."\
                 +" We added missing samples with area value zero in raw intensity file "
            return {'error': True, 'msg': [error_msg]}

        return {'error': False, 'msg': msgs_list}

    def _merge_mq_metadata(self):
        """
        merge the mapping of fragments to unlabeled fragment
        with the raw mq dataframe
        """
        assert set(self.metadata_mq.get(constants.PARENT)).issubset(
            set(self.metadata_mq[constants.MQ_FRAGMENT])), \
            str(constants.MQ_FRAGMENT + ' must be subset of ' + constants.PARENT)
        self.raw_mq = self.raw_mq.merge(self.metadata_mq, how='inner',
                                        on=constants.MQ_FRAGMENT)
        if self.raw_mq.empty:
            raise Exception(err_cnsts.INPUTERR5_UNCOMMON_ENTRIES.format(
                            constants.MQ_FRAGMENT))
        self.raw_mq.fillna(0, inplace=True)


    def _split_std_bio(self):
        """
        Divide the raw df into standard and biosample

        Returns:
            standard_df: part of raw df containing standard samples
            biosamples_df: part of raw df containing biosamples, with a conc
            column added to it containing -1s
        """
        std_sample_list = self.metadata_std[constants.MQ_SAMPLE_NAME].unique()
        standard_df = self.raw_mq[self.raw_mq[constants.MQ_SAMPLE_NAME].isin(
            std_sample_list)]

        if standard_df.empty:
            raise ValueError(
                err_cnsts.INPUTERR18_NO_STD_SAMPLE_MQ_COH)
        biosample_index = self.raw_mq.index.difference(standard_df.index)
        biosamples_df = self.raw_mq.loc[biosample_index, :]
        if biosamples_df.empty:
            raise ValueError(
                err_cnsts.INPUTERR19_NO_BIO_SAMPLE_MQ_COH)

        # compute sum by fragments
        # TODO: make a utility function, used here and in tableprocessing
        intensity_sum = \
            biosamples_df.groupby([constants.PARENT, constants.MQ_SAMPLE_NAME])[
                constants.MQ_INTENSITY].sum().reset_index()
        del biosamples_df[constants.MQ_INTENSITY]
        biosamples_df = biosamples_df.merge(intensity_sum, how="inner",
                                            left_on=[constants.PARENT,
                                                     constants.MQ_SAMPLE_NAME],
                                            right_on=[constants.PARENT,
                                                      constants.MQ_SAMPLE_NAME])
        biosamples_df.reset_index(inplace=True)
        biosamples_df = biosamples_df.drop_duplicates(
            subset=[constants.PARENT, constants.MQ_SAMPLE_NAME])

        biosamples_df[constants.MQ_CONCENTRATION] = pd.Series(-1,
                                                              biosamples_df.index)
        return standard_df, biosamples_df

    def _merge_std_df_metadata(self, standard_df):
        """
        merge concentration mappings to standard dataframe

        Args:
            standard_df: df containing std samples

        Returns:
            merged_std: df containing std samples mapped to corresponding conc
        """
        merged_std = standard_df.merge(self.metadata_std, how='inner',
                                       on=[constants.MQ_SAMPLE_NAME,
                                           constants.MQ_COHORT_NAME,
                                           constants.PARENT])

        if merged_std.empty:
            cols = constants.MQ_SAMPLE_NAME + ',' + constants.MQ_COHORT_NAME + ' or ' + constants.PARENT
            raise Exception(err_cnsts.INPUTERR5_UNCOMMON_ENTRIES.format(cols, 'metadata_std', 'raw_files'))
        return merged_std


def process_input_data(input_local_files, normalisation_with_intensity,
                       internal_std_name, isMetadataPresent, edited_data,
                       sample_name):
    """
    This function validate and process input files and returns merged_df,
    biosample and logs

    Args:
        input_local_files : All three files required as input for tool
        normalisation_with_intensity: Normalisation Intenity value
        internal_std_name: internal standard Name values
        isMetadataPresent: boolean to check if metadata file is present
        edited_data: list of dictionaries to append to metadata.
        sample_name: name of sample that user want to use as max intensity
        sample to normalize

    Returns:
        merged_df: standard sample dataframe with corresponding
                    concentration values
        biosample: biosample dataframe
        logs: logs in raw file
        missing_comp_logs: Missing component logs
    """

    mq_files, raw_file_logs = read_raw_file(input_local_files.get(
        constants.MQ_FILES)[0])
    mq_files, neg_logs = dat_hlp.replace_neg_with_zero(mq_files, [constants.MQ_INTENSITY])
    for error_msg in neg_logs:
        raw_file_logs['warnings']['message'].append(error_msg)

    mq_sample_metadata, sample_metadata_logs = validate_via_olmonk(
        constants.DICT_READ_STD_SAMPLE_METADATA,
        input_local_files.get(constants.METADATA_STD_SAMPLE_FILE)[0])
    try:
        mq_sample_metadata = mq_sample_metadata.drop(['Area'], axis=1)
    except:
        pass

    log_data = find_missing_samples(mq_sample_metadata, mq_files)

    if log_data:
        sample_metadata_logs[constants.ERRORS]= log_data['msg']

    if isMetadataPresent == 1:
        missing_comp_logs = {}
        missing_comp_logs[constants.MISSING_COMPONENTS] = []
        mq_metadata, metadata_mq_logs = df_csv_validator(
            constants.DICT_READ_MULTIQUANT_METADATA,
            input_local_files.get(constants.METADATA_MQ_FILE)[0])


    elif isMetadataPresent == 0 and not raw_file_logs[constants.ERRORS] and\
    dat_hlp.check_df_type(mq_files) == 'ms':
        metadata_mq_logs = {'errors': [] , 'warnings': {'action': [], 'message': []}}
        mq_metadata, missing_comp_logs = dat_alg.quant_create_fragment_mapping(
            mq_files, edited_data)

    else:
        metadata_mq_logs = {'errors': [], 'warnings': {'action': [], 'message': []}}
        mq_metadata = meta_gen.create_metadata_mq(mq_files)
        missing_comp_logs = {}
        missing_comp_logs[constants.MISSING_COMPONENTS] = []

    if raw_file_logs[constants.ERRORS] or metadata_mq_logs[constants.ERRORS]\
        or sample_metadata_logs[constants.ERRORS]:
        return None, None, raw_file_logs, sample_metadata_logs, metadata_mq_logs,\
               [], None

    # calling normalisation wrt to the parameters provided by user.
    merged_std, biosample, outlier_data, raw_file_logs = normalise(mq_files,
        mq_sample_metadata, mq_metadata, internal_std_name,
        normalisation_with_intensity, sample_name, raw_file_logs,
        log_data)

    return merged_std, biosample, raw_file_logs, sample_metadata_logs, \
            metadata_mq_logs, missing_comp_logs, outlier_data


def normalise(mq_files, mq_sample_metadata, mq_metadata, internal_std_name,
     normalisation_with_intensity, sample_name, raw_file_logs, log_data):
    """
    This function normalises the intensities based
    on the selected internal standard by the user.

    Parameters
    ----------
        mq_files: dataframe
            merged output of raw file and metadata
        mq_sample_metadatda: dataframe
            sample metadata
        mq_metadata: dataframe
            fragment and unlabeled fragment mapping
        internal_std_name: string
            name of the sample which needs to be normalised against
        sample_name: string
            name of the sample
        raw_file_logs: dict
            logs of errors and warnings of raw_file


    returns
    --------
        merged_std: dataframe
            merged output of raw file and metadata normalised
    """
    try:
        if normalisation_with_intensity:
            merged_std, biosample= nl.get_normalized_biosample_data(mq_files,
                                                            mq_sample_metadata,
                                                            mq_metadata,
                                                            internal_std_name, sample_name)
            #missing_comp_logs[constants.MISSING_COMPONENTS] = []
            outlier_data = []
        else:
            process_mq_df = ProcessMqDf(mq_files, mq_sample_metadata, mq_metadata)
            # warn_msg = process_mq_df.check_std_samples_in_rawfile()
            # if log_data:
            #     raw_file_logs[constants.WARNINGS][constants.MESSAGE] = warn_msg['msg']
            merged_std, biosample = process_mq_df.get_standard_biosample_data()
            outlier_data = nl.get_outliers_data(mq_files, mq_metadata)
    except Exception as e:
        merged_std, biosample = pd.DataFrame(), pd.DataFrame()
        outlier_data = []
        raw_file_logs[constants.ERRORS].append(str(e))
        raw_file_logs[constants.WARNINGS] = {constants.ACTION: [], constants.MESSAGE: []}
    return merged_std, biosample, outlier_data, raw_file_logs


def get_standard_data(standard_table, metabolite, std_conc_col):
    """
    Returns Standard sample for the corresponding metabolite .

    Args:
        standard_table: standard sample df
        metabolite: metabolite name, for which standard table is created
        std_conc_col: concentrations of std samples

    Returns:
            internal_data: Standard Data for a fragment as a json string.
    """
    internal_data = []
    for index, row in standard_table[standard_table[constants.PARENT] == metabolite].iterrows():
        areakey = []
        areavalue = []
        for column_name in standard_table.columns.values:
            if 'Area_' in column_name:
                areakey.append(str(column_name))
                areavalue.append(float('%.6f' % (row[column_name])))
        areavalue = ["" if val < 0 else val for val in areavalue]
        internal_data.append({constants.CONC: row[std_conc_col],
                              constants.AREA_KEY: areakey,
                              constants.AREA_VALUE: areavalue,
                              constants.AVGAREA: row[constants.AVG_AREA],
                              constants.STDDEV: row[constants.STD_DEV]})
    return internal_data


def create_read_input_file_response_obj(merged_std, biosample, background_corr):
    """
    Returns standard data of each metabolite in the biosample and warnings
    Args:
        merged_std: dataframe containing all standard samples with
                    corresponding concentration.
        biosamples: dataframe containing experiment samples, for which
                    standard samples are needed.
        background_corr: boolean values corresponding to background correction
    Returns:
        upload_info: standard_data for each metabolite in the biosample
                and the warning message.
    """
    warn_message = ''
    table_processing = TableProcessing(merged_std)
    with wrngs.catch_warnings(record=True) as w:
        wrngs.filterwarnings('always', category=UserWarning)
        standard_table, y_unit = table_processing.get_all_std_tables_df(
            background_corr)

    if len(w) == 1:
        warn_message = w[0].message

    metabolites_df = biosample[constants.PARENT].unique()

    metabs = []
    for metabolite in metabolites_df:
        metabs.append({constants.NAME: metabolite,
                       constants.STD_DATA: get_standard_data(standard_table,
                                                             metabolite,
                                                    table_processing.conc_col)})
    upload_info = {constants.METABOLITE_INFO: metabs,
                   constants.WARNING: str(warn_message)}

    return upload_info


def find_missing_samples(sample_df,raw_df):
    """
    This function checks whether any sample in the standard sample file are
    missing in the raw intensity file

    Args:
    sample_df: dataframe of standard sample file
    raw_df : dataframe of raw intensity file

    Returns:
    dict containing error status and error message (if any)
    """
    missing_sample_dict = {}
    msgs_list = []
    standard_msg_list = []
    raw_samples_set = set(raw_df[constants.MQ_SAMPLE_NAME])
    standard_sample_set = set(sample_df[constants.MQ_SAMPLE_NAME])
    missing_samples = standard_sample_set - raw_samples_set
    if len(missing_samples):
        error_msg = "Sample {} ".format(','.join(str(data)for data in missing_samples))\
         +"in your standard sample metadata are missing in the raw intensity file. "\
         +"Please check them and retry."
        return {'error': True, 'msg' : [error_msg]}
    return {'error': False, 'msg': msgs_list}
